#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"
#include "nrf_delay.h"
#include "nrf_drv_gpiote.h"
#include "nrf_drv_twi.h"
#include "bno080-nrf-twi.h"

#define TWI_INSTANCE_ID     0
#define TWI_ADDRESSES       127

//SH-2 - Header Packet Construct
typedef struct
{
    uint8_t     length_lsb;
    uint8_t     length_msb;
    uint8_t     channel;
    uint8_t     seqNum;
    uint16_t    length;
} m_sh2_header;

m_sh2_header    shtpHeader; //Each packet has a header of 4 bytes
uint8_t         shtpData[MAX_PACKET_SIZE];
uint8_t         sequenceNumber[6] = {0, 0, 0, 0, 0, 0}; //There are 6 com channels. Each channel has its own seqnum
uint8_t         commandSequenceNumber = 0;				//Commands have a seqNum as well. These are inside command packet, the header uses its own seqNum per channel

static const nrf_drv_twi_t m_twi_instance = NRF_DRV_TWI_INSTANCE(TWI_INSTANCE_ID);
volatile static bool twi_tx_done = false;
volatile static bool twi_rx_done = false;

uint8_t twi_tx_buffer[MPU_TWI_BUFFER_SIZE];

/**
 * @brief TWI Event Handler.
 * Just the usual way. Nothing special here
 */
static void nrf_drv_mpu_twi_event_handler(nrf_drv_twi_evt_t const * p_event, void * p_context)
{
    switch(p_event->type)
    {
        case NRF_DRV_TWI_EVT_DONE:
            NRF_LOG_INFO ("NRF_DRV_TWI_EVT_DONE");
            switch(p_event->xfer_desc.type)
            {
                case NRF_DRV_TWI_XFER_TX:
                    //NRF_LOG_INFO ("NRF_DRV_TWI_XFER_TX | Bytes Sent : %d", nrf_drv_twi_data_count_get(&m_twi_instance) );
                    twi_tx_done = true;
                    break;
                case NRF_DRV_TWI_XFER_TXTX:
                    //NRF_LOG_INFO ("NRF_DRV_TWI_XFER_TXTX");
                    twi_tx_done = true;
                    break;
                case NRF_DRV_TWI_XFER_RX:
                    //NRF_LOG_INFO ("NRF_DRV_TWI_XFER_RX");
                    twi_rx_done = true;
                    break;
                case NRF_DRV_TWI_XFER_TXRX:
                    //NRF_LOG_INFO ("NRF_DRV_TWI_XFER_TXRX");
                    twi_rx_done = true;
                    break;
                default:
                    //NRF_LOG_INFO ("NRF_DRV_TWI_EVT_DONE - EVT UNKNOWN");
                    break;
            }
            break;
        case NRF_DRV_TWI_EVT_ADDRESS_NACK:
            NRF_LOG_INFO ("NRF_DRV_TWI_EVT_ADDRESS_NACK");
            break;
        case NRF_DRV_TWI_EVT_DATA_NACK:
            NRF_LOG_INFO ("NRF_DRV_TWI_EVT_DATA_NACK");
            break;
        default:
            break;
    }
}

/**
 * @brief TWI initialization.
 * Interrupt PIN Handler
 */
void bno080_in_pin_handler()
{
    //NRF_LOG_INFO("INT PIN Toggled - BNO080 : %d", nrf_gpio_pin_read(MPU_INT_PIN));
    shtp_get_header();
}

/**
 * @brief BNO080 Reset.
 * Soft Reset of the BNO080
 */
bool shtp_get_header()
{
    uint32_t err_code;

    //uint32_t timeout = MPU_TWI_TIMEOUT;
    //while( (!twi_tx_done) && (!twi_rx_done) && --timeout);
    //if(!timeout) return NRF_ERROR_BUSY;
    //reset the flags
    //twi_tx_done = false;
    //twi_rx_done = false;

    //Header Read
    uint16_t packetLength = 4;
    uint8_t * header_buffer = malloc(packetLength);
    nrf_drv_twi_xfer_desc_t xfer_desc;
    xfer_desc.type = NRF_DRV_TWI_XFER_RX;
    xfer_desc.address = BNO080_TWI_ADDRESS;
    xfer_desc.primary_length = packetLength;
    xfer_desc.p_primary_buf = header_buffer;
    err_code = nrf_drv_twi_xfer(&m_twi_instance, &xfer_desc, 0);

    //Calculate the number of data bytes in this packet
    int16_t dataLength = ((uint16_t)header_buffer[1] << 8 | header_buffer[0]);
	dataLength &= ~(1 << 15); //Clear the MSbit.

    shtpHeader.length_lsb  = header_buffer[0];
    shtpHeader.length_msb  = header_buffer[1];
    shtpHeader.length      = dataLength;
    shtpHeader.channel     = header_buffer[2];
    shtpHeader.seqNum      = header_buffer[3];
    
    //NRF_LOG_INFO("SH-2 Header 0 - 1 - 2 - 3 : 0x%x 0x%x 0x%x 0x%x", shtpHeader.length_lsb, shtpHeader.length_msb, shtpHeader.channel, shtpHeader.seqNum);
    NRF_LOG_INFO("SH-2 Header Bytes : %d", shtpHeader.length);
    NRF_LOG_FLUSH();

    if (shtpHeader.length > 4){return (false);} //Packet is empty - All done

    return true;

    //Read all incoming data and flush it
	//nrf_delay_ms(50);
	//while (bno080_read() == true);
    //bno080_read();

}

/**
 * @brief TWI initialization.
 * Just the usual way. Nothing special here
 */
uint32_t nrf_drv_mpu_init(void)
{
    uint32_t err_code;
    
    const nrf_drv_twi_config_t twi_mpu_config = {
       .scl                = TWI_SCL_PIN,
       .sda                = TWI_SDA_PIN,
       .frequency          = NRF_TWI_FREQ_100K,
       .interrupt_priority = APP_IRQ_PRIORITY_HIGHEST,
       .clear_bus_init     = true
    };
    
    err_code = nrf_drv_twi_init(&m_twi_instance, &twi_mpu_config, nrf_drv_mpu_twi_event_handler, NULL);
    if(err_code != NRF_SUCCESS)
	{
		return err_code;
	}
    
    nrf_drv_twi_enable(&m_twi_instance);

    NRF_LOG_INFO ("nrf_drv_mpu_init");
	
	return NRF_SUCCESS;
}

/**
 * @brief BNO080 Reset.
 * Soft Reset of the BNO080
 */
uint32_t bno080_softreset()
{
    uint32_t err_code;
    uint32_t timeout = MPU_TWI_TIMEOUT;

    while((mpu_interrupt) && (twi_tx_done) && --timeout);
    if(!timeout) return NRF_ERROR_BUSY;

    uint16_t packetLength = 5;
    uint8_t * txbuffer = malloc(packetLength);          //setup the data paxket to send
    txbuffer[0] = packetLength & 0xFF;                  //SHTP Header - Byte 0 - Packet Length LSB
    txbuffer[1] = packetLength >> 8;                    //SHTP Header - Byte 1 - Packet Length MSB
    txbuffer[2] = CHANNEL_EXECUTABLE;                   //SHTP Header - Byte 2 - Channel Number
    txbuffer[3] = sequenceNumber[CHANNEL_EXECUTABLE]++; //SHTP Header - Byte 3 - Sequence Number
    txbuffer[4] = 0x01;                                 //SHTP Header - Byte 4 - SoftReset

    // Setting up transfer
    nrf_drv_twi_xfer_desc_t xfer_desc;
    xfer_desc.type = NRF_DRV_TWI_XFER_TX;
    xfer_desc.address = BNO080_TWI_ADDRESS;
    xfer_desc.primary_length = packetLength;
    xfer_desc.p_primary_buf = txbuffer;  

    // Transferring
    err_code = nrf_drv_twi_xfer(&m_twi_instance, &xfer_desc, 0);

    return err_code;
}

/**
 * @brief BNO080 Scanner.
 * Scan the TWI bus for the BNO080
 */
uint32_t  bno080_init()
{
    // This burst write function is not optimal and needs improvement.
    // The new SDK 11 TWI driver is not able to do two transmits without repeating the ADDRESS + Write bit byte
    uint32_t err_code;
    uint32_t timeout = MPU_TWI_TIMEOUT;

    while((mpu_interrupt) && (twi_tx_done) && --timeout);
    if(!timeout) return NRF_ERROR_BUSY;

    uint16_t packetLength = 6;
    uint8_t * txbuffer = malloc(packetLength);          //setup the data paxket to send
    txbuffer[0] = packetLength & 0xFF;                  //SHTP Header - Byte 0 - Packet Length LSB
    txbuffer[1] = packetLength >> 8;                    //SHTP Header - Byte 1 - Packet Length MSB
    txbuffer[2] = CHANNEL_CONTROL;                      //SHTP Header - Byte 2 - Channel Number
    txbuffer[3] = sequenceNumber[CHANNEL_CONTROL]++;    //SHTP Header - Byte 3 - Sequence Number
    txbuffer[4] = SHTP_REPORT_PRODUCT_ID_REQUEST;       //SHTP Header - Byte 4 - SHTP_REPORT_PRODUCT_ID_REQUEST
    txbuffer[5] = 0x0;                                  //SHTP Header - Byte 5 - Reserved
     
    // Setting up transfer
    nrf_drv_twi_xfer_desc_t xfer_desc;
    xfer_desc.address = BNO080_TWI_ADDRESS;
    xfer_desc.type = NRF_DRV_TWI_XFER_TX;
    xfer_desc.primary_length = packetLength;
    xfer_desc.p_primary_buf = txbuffer;  

    // Transferring
    err_code = nrf_drv_twi_xfer(&m_twi_instance, &xfer_desc, 0);

    return err_code;
}

/**
 * @brief BNO080 Scanner.
 * Scan the TWI bus for the BNO080
 */
uint32_t twi_scan_for_bno080()
{

    uint32_t err_code;
    bool detected_device = false;
    uint32_t timeout = MPU_TWI_TIMEOUT;

    while((mpu_interrupt) && (twi_tx_done) && --timeout);
    if(!timeout) return NRF_ERROR_BUSY;

    uint16_t packetLength = 1;
    uint8_t * txbuffer = malloc(packetLength);          //setup the data paxket to send
    txbuffer[0] = 0;
     
    // Setting up transfer
    nrf_drv_twi_xfer_desc_t xfer_desc;
    xfer_desc.address = BNO080_TWI_ADDRESS;
    xfer_desc.type = NRF_DRV_TWI_XFER_TX;
    xfer_desc.primary_length = packetLength;
    xfer_desc.p_primary_buf = txbuffer;  

    // Transferring
    err_code = nrf_drv_twi_xfer(&m_twi_instance, &xfer_desc, NRF_DRV_TWI_FLAG_NO_XFER_EVT_HANDLER);

    if (err_code == NRF_SUCCESS)
    {
        detected_device = true;
        NRF_LOG_INFO("BNO080 device detected at address 0x%x.", BNO080_TWI_ADDRESS);
    }
    NRF_LOG_FLUSH();

    if (!detected_device)
    {
        NRF_LOG_INFO("No device was found.");
        return NRF_ERROR_INVALID_ADDR;
    }

    return NRF_SUCCESS;
}

/**
 * @brief TWI Scanner.
 * Just the usual way. Nothing special here
 */
uint32_t twi_scanner()
{

    uint32_t err_code;
    uint8_t address;
    uint8_t sample_data;
    bool detected_device = false;

    NRF_LOG_INFO("twi_scanner");

    for (address = 1; address <= TWI_ADDRESSES; address++)
    {
        err_code = nrf_drv_twi_rx(&m_twi_instance, address, &sample_data, sizeof(sample_data));
        if (err_code == NRF_SUCCESS)
        {
            detected_device = true;
            NRF_LOG_INFO("TWI device detected at address 0x%x.", address);
        }
        NRF_LOG_FLUSH();
    }

    if (!detected_device)
    {
        NRF_LOG_INFO("No device was found.");
        return NRF_ERROR_INVALID_ADDR;
    }

    return NRF_SUCCESS;

}